import logging
import shutil
from os import path



import test.mock
from test import mock
import autofit as af
import autofit as af

logger = logging.getLogger(__name__)

try:
    output_dir = "{}/../../workspace/output".format(path.dirname(path.realpath(__file__)))
    logger.info("Removing {}".format(output_dir))
    shutil.rmtree(output_dir)
except FileNotFoundError:
    logging.info("Not found")


class Analysis(af.Analysis):
    def fit(self, instance):
        return -(instance.profile.centre[0] ** 2 + instance.profile.centre[1] ** 2)

    def visualize(self, instance, **kwargs):
        pass

    def describe(self, instance):
        return "{}, {}".format(*instance.profile.centre)


if __name__ == "__main__":
    grid_search = gs.GridSearch(phase_name="phase_grid_search", phase_tag='_tag', phase_folders=['integration'],
                                optimizer_class=af.MultiNest, parallel=True)
    grid_search.variable.profile = test.mock.EllipticalProfile

    # noinspection PyUnresolvedReferences
    result = grid_search.fit(Analysis(),
                             [grid_search.variable.profile.centre_0, grid_search.variable.profile.centre_1])

    print(result)



from .non_linear.downhill_simplex import DownhillSimplex
from .non_linear.grid_search import GridSearch
from .non_linear.multi_nest import MultiNest
from .non_linear.non_linear import Analysis
from .non_linear.non_linear import NonLinearOptimizer
from .non_linear.non_linear import Result



import inspect

from autofit.mapper.prior import Prior
from autofit.mapper.model import AbstractModel
from autofit.mapper.prior import cast_collection, PriorNameValue, ConstantNameValue
from autofit.mapper.prior_model.util import PriorModelNameValue


class AbstractPriorModel(AbstractModel):
    """
    Abstract model that maps a set of priors to a particular class. Must be
    overridden by any prior model so that the model mapper recognises its prior \
    model attributes.

    @DynamicAttrs
    """

    @property
    def name(self):
        return self.__class__.__name__

    # noinspection PyUnusedLocal
    @staticmethod
    def from_object(t, *args, **kwargs):
        if inspect.isclass(t):
            from .prior_model import PriorModel
            obj = object.__new__(PriorModel)
            obj.__init__(t, **kwargs)
        elif isinstance(t, list) or isinstance(t, dict):
            from autofit.mapper import CollectionPriorModel
            obj = object.__new__(CollectionPriorModel)
            obj.__init__(t)
        else:
            obj = t
        return obj

    @property
    def info(self):
        info = []

        prior_model_iterator = self.direct_prior_tuples + self.direct_constant_tuples

        for attribute_tuple in prior_model_iterator:
            attribute = attribute_tuple[1]

            # noinspection PyUnresolvedReferences
            line = attribute_tuple.name
            # noinspection PyUnresolvedReferences
            info.append(line + ' ' * (60 - len(line)) + str(attribute))

        for prior_model_name, prior_model in self.prior_model_tuples:
            info.append(prior_model.name + '\n')
            info.extend([f"{prior_model_name}_{item}" for item in prior_model.info])

        return info

    @property
    @cast_collection(PriorNameValue)
    def direct_prior_tuples(self):
        return self.tuples_with_type(Prior)

    @property
    @cast_collection(ConstantNameValue)
    def direct_constant_tuples(self):
        return self.tuples_with_type(float)

    @property
    def flat_prior_model_tuples(self):
        """
        Returns
        -------
        prior_models: [(str, AbstractPriorModel)]
            A list of prior models associated with this instance
        """
        raise NotImplementedError(
            "PriorModels must implement the flat_prior_models property")

    @property
    @cast_collection(PriorModelNameValue)
    def prior_model_tuples(self):
        return self.tuples_with_type(AbstractPriorModel)

    @property
    def prior_tuples(self):
        raise NotImplementedError()

    @property
    @cast_collection(PriorModelNameValue)
    def direct_prior_model_tuples(self):
        return [(name, value) for name, value in self.__dict__.items() if
                isinstance(value, AbstractPriorModel)]

    def __eq__(self, other):
        return isinstance(other, AbstractPriorModel) \
               and self.direct_prior_model_tuples == other.direct_prior_model_tuples

    @property
    def constant_tuples(self):
        raise NotImplementedError()

    @property
    def prior_class_dict(self):
        raise NotImplementedError()

    def instance_for_arguments(self, arguments):
        raise NotImplementedError()

    @property
    def prior_count(self):
        return len(self.prior_tuples)

    @property
    def priors(self):
        return [prior_tuple.prior for prior_tuple in self.prior_tuples]

    def name_for_prior(self, prior):
        for prior_model_name, prior_model in self.direct_prior_model_tuples:
            prior_name = prior_model.name_for_prior(prior)
            if prior_name is not None:
                return "{}_{}".format(prior_model_name, prior_name)
        prior_tuples = self.prior_tuples
        for name, p in prior_tuples:
            if p == prior:
                return name

    def __hash__(self):
        return self.id



from .abstract import AbstractPriorModel
from .annotation import AnnotationPriorModel
from .collection import CollectionPriorModel
from .deferred import DeferredArgument
from .deferred import DeferredInstance
from .prior_model import PriorModel
from .util import PriorModelNameValue
from .util import is_tuple_like_attribute_name
from .util import tuple_name



from autofit import exc


class DeferredInstance:
    def __init__(
            self,
            cls: type,
            constructor_arguments: {str: object}
    ):
        """
        An instance that has been deferred for later construction

        Parameters
        ----------
        cls
            The class to be constructed
        constructor_arguments
            The arguments provided by the optimiser
        """
        self.cls = cls
        self.constructor_arguments = constructor_arguments

    @property
    def deferred_argument_names(self) -> [str]:
        """
        The names of arguments still required to instantiate the class
        """
        return [
            name for name, value
            in self.constructor_arguments.items()
            if isinstance(value, DeferredArgument)
        ]

    def __call__(self, **kwargs):
        """
        Constructs an instance of the class provided that all unset arguments are
        passed.

        Parameters
        ----------
        kwargs
            Key value pairs for arguments that should be set

        Returns
        -------
        instance: self.cls
            An instance of the class
        """
        return self.cls(**{**self.constructor_arguments, **kwargs})

    def __getattr__(self, item):
        """
        Failing to get an attribute is considered to indicate an attempt to use a
        deferred instance without first instantiating it. As such an exception is
        raised to warn the user that they need to instantiate the class.
        """
        try:
            super().__getattribute__(item)
        except AttributeError:
            raise exc.DeferredInstanceException(
                f"{self.cls.__name__} cannot be called until it is instantiated with"
                f" deferred arguments {self.deferred_argument_names}"
            )


class DeferredArgument(object):
    """
    A deferred argument which is passed into the construct the final instance after
    model mapper instance generation
    """
    pass


class PriorException(Exception):
    pass


class MultiNestException(Exception):
    pass


class CheckpointException(Exception):
    """
    Thrown when a non linear optimiser recovery fails due to assertion failure
    """
    pass


class FitException(Exception):
    """
    An exception to be thrown if the non linear search must resample; equivalent to returning an infinitely bad fit
    """
    pass


class PriorLimitException(FitException, PriorException):
    pass


class PipelineException(Exception):
    pass


class DeferredInstanceException(Exception):
    """
    Exception raised when an attempt is made to access an attribute or function of a
    deferred instance prior to instantiation
    """
    pass



import os
import shutil

import autofit as af

text_path = "{}/../test_files/text/".format(os.path.dirname(os.path.realpath(__file__)))


def test__label_and_label_string():
    string0 = af.text_util.label_and_label_string(label0='param0', label1='mass',
                                               whitespace=10)
    string1 = af.text_util.label_and_label_string(label0='param00', label1='mass0',
                                               whitespace=10)
    string2 = af.text_util.label_and_label_string(label0='param000', label1='mass111',
                                               whitespace=10)

    assert string0 == 'param0    mass'
    assert string1 == 'param00   mass0'
    assert string2 == 'param000  mass111'

    string0 = af.text_util.label_and_label_string(label0='param0', label1='mass',
                                               whitespace=20)

    assert string0 == 'param0              mass'


def test__label_and_value_string():
    string0 = af.text_util.label_and_value_string(label='param0', value=2.0, whitespace=10)
    string1 = af.text_util.label_and_value_string(label='param00', value=2.0,
                                               whitespace=10)
    string2 = af.text_util.label_and_value_string(label='param000', value=2.0,
                                               whitespace=10)

    assert string0 == 'param0    2.00'
    assert string1 == 'param00   2.00'
    assert string2 == 'param000  2.00'

    string = af.text_util.label_and_value_string(label='param11', value=3.00,
                                              whitespace=20)

    assert string == 'param11             3.0000'

    string = af.text_util.label_and_value_string(label='param12', value=3.00,
                                              whitespace=15)

    assert string == 'param12        3.00e+00'


def test__label_value_and_limits_string():
    string0 = af.text_util.label_value_and_limits_string(label='param0', value=2.0,
                                                      upper_limit=2.5, lower_limit=1.5,
                                                      whitespace=10)
    string1 = af.text_util.label_value_and_limits_string(label='param00', value=2.0,
                                                      upper_limit=2.7, lower_limit=1.3,
                                                      whitespace=10)
    string2 = af.text_util.label_value_and_limits_string(label='param000', value=2.0,
                                                      upper_limit=2.9, lower_limit=1.1,
                                                      whitespace=10)

    assert string0 == 'param0    2.00 (1.50, 2.50)'
    assert string1 == 'param00   2.00 (1.30, 2.70)'
    assert string2 == 'param000  2.00 (1.10, 2.90)'

    string = af.text_util.label_value_and_limits_string(label='param11', value=3.00,
                                                     upper_limit=40000.0,
                                                     lower_limit=0.0001,
                                                     whitespace=20)

    assert string == 'param11             3.0000 (0.0001, 40000.0000)'

    string = af.text_util.label_value_and_limits_string(label='param12', value=3.00,
                                                     upper_limit=500.0, lower_limit=1.0,
                                                     whitespace=15)

    assert string == 'param12        3.00e+00 (1.00e+00, 5.00e+02)'


def test__label_value_and_unit_string():
    string0 = af.text_util.label_value_and_unit_string(label='param0', value=2.0,
                                                    unit='arcsec', whitespace=10)
    string1 = af.text_util.label_value_and_unit_string(label='param00', value=2.0,
                                                    unit='mass', whitespace=10)
    string2 = af.text_util.label_value_and_unit_string(label='param000', value=2.0,
                                                    unit='kg', whitespace=10)

    assert string0 == 'param0    2.00 arcsec'
    assert string1 == 'param00   2.00 mass'
    assert string2 == 'param000  2.00 kg'

    string = af.text_util.label_value_and_unit_string(label='param11', value=3.00,
                                                   unit='kg', whitespace=20)

    assert string == 'param11             3.0000 kg'

    string = af.text_util.label_value_and_unit_string(label='param12', value=3.00,
                                                   unit='kgs', whitespace=15)

    assert string == 'param12        3.00e+00 kgs'


def test__output_list_of_strings_to_file():
    if os.path.exists(text_path):
        shutil.rmtree(text_path)

    os.mkdir(text_path)

    results = ['hi\n', 'hello']
    af.text_util.output_list_of_strings_to_file(file=text_path + 'model.results',
                                             list_of_strings=results)

    file = open(text_path + 'model.results', 'r')

    assert file.readlines() == ['hi\n', 'hello']



import os
import pickle

import autofit.optimize.non_linear.multi_nest
import autofit.optimize.non_linear.non_linear
from autofit import conf
from autofit import exc
from autofit.optimize import grid_search


class HyperPhase(object):
    pass


class AbstractPhase(object):

    def __init__(
            self,
            phase_name,
            tag_phases=True,
            phase_tag=None,
            phase_folders=tuple(),
            optimizer_class=autofit.optimize.non_linear.multi_nest.MultiNest,
            auto_link_priors=False
    ):
        """
        A phase in an lensing pipeline. Uses the set non_linear optimizer to try to
        fit_normal models and image passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        phase_name: str
            The name of this phase
        """

        self.tag_phases = tag_phases

        if phase_tag is None and tag_phases:
            self.phase_tag = ''
        else:
            self.phase_tag = 'settings' + phase_tag

        self.optimizer = optimizer_class(
            phase_name=phase_name,
            phase_tag=phase_tag,
            phase_folders=phase_folders
        )
        self.auto_link_priors = auto_link_priors

    @property
    def phase_path(self):
        return self.optimizer.phase_path

    @phase_path.setter
    def phase_path(self, phase_path):
        self.optimizer.phase_path = phase_path

    @property
    def phase_name(self):
        return self.optimizer.phase_name

    @phase_name.setter
    def phase_name(self, phase_name):
        self.optimizer.phase_name = phase_name

    @property
    def variable(self):
        """
        Convenience method

        Returns
        -------
        ModelMapper
            A model mapper comprising all the variable (prior) objects in this lensing
        """
        return self.optimizer.variable

    def run_analysis(self, analysis):
        return self.optimizer.fit(analysis)

    @property
    def path(self):
        return self.optimizer.path

    @property
    def doc(self):
        if self.__doc__ is not None:
            return self.__doc__.replace("  ", "").replace("\n", " ")

    def pass_priors(self, results):
        """
        Perform any prior or constant passing. This could involve setting model
        attributes equal to priors or constants from a previous phase.

        Parameters
        ----------
        results: ResultsCollection
            The result of the previous phase
        """
        pass

    # noinspection PyAbstractClass
    class Analysis(autofit.optimize.non_linear.non_linear.Analysis):

        def __init__(self, results=None):
            """
            An lensing object

            Parameters
            ----------
            results: ResultsCollection
                The results of all previous phases
            """

            self.results = results

        @property
        def last_results(self):
            if self.results is not None:
                return self.results.last

    def make_result(self, result, analysis):
        raise NotImplementedError()

    def make_optimizer_pickle_path(self) -> str:
        """
        Create the path at which the optimizer pickle should be saved
        """
        return "{}/.optimizer.pickle".format(self.make_path())

    def make_path(self) -> str:
        """
        Create the path to the folder at which the metadata and optimizer pickle should
        be saved
        """
        return "{}/{}{}/{}".format(conf.instance.output_path, self.phase_path,
                                   self.phase_name, self.phase_tag)

    def save_optimizer_for_phase(self):
        """
        Save the optimizer associated with the phase as a pickle
        """
        with open(self.make_optimizer_pickle_path(), "w+b") as f:
            f.write(pickle.dumps(self.optimizer))

    def save_metadata(self, data_name, pipeline_name):
        """
        Save metadata associated with the phase, such as the name of the pipeline, the
        name of the phase and the name of the data being fit
        """
        with open("{}/.metadata".format(self.make_path()), "w+") as f:
            f.write(
                "pipeline={}\nphase={}\ndata={}".format(pipeline_name, self.phase_name,
                                                        data_name))

    def assert_optimizer_pickle_matches_for_phase(self):
        """
        Assert that the previously saved optimizer is equal to the phase's optimizer if
        a saved optimizer is found.

        Raises
        -------
        exc.PipelineException
        """
        path = self.make_optimizer_pickle_path()
        if os.path.exists(path):
            with open(path, "r+b") as f:
                loaded_optimizer = pickle.loads(f.read())
                if self.optimizer != loaded_optimizer:
                    raise exc.PipelineException(
                        f"Can't restart phase at path {path} because settings don't "
                        f"match. Did you change the optimizer settings or model?")

    def assert_and_save_pickle(self):
        if conf.instance.general.get("output", "assert_pickle_matches", bool):
            self.assert_optimizer_pickle_matches_for_phase()
        self.save_optimizer_for_phase()


def as_grid_search(phase_class, parallel=False):
    """
    Create a grid search phase class from a regular phase class. Instead of the phase
    being optimised by a single non-linear optimiser, a new optimiser is created for
    each square in a grid.

    Parameters
    ----------
    phase_class
        The original phase class
    parallel: bool
        Indicates whether non linear searches in the grid should be performed on
        parallel processes.

    Returns
    -------
    grid_search_phase_class: GridSearchExtension
        A class that inherits from the original class, replacing the optimiser with a
        grid search optimiser.

    """

    class GridSearchExtension(phase_class):
        def __init__(self, *args, phase_name, tag_phases=True, phase_folders=tuple(),
                     number_of_steps=10,
                     optimizer_class=autofit.optimize.non_linear.multi_nest.MultiNest,
                     **kwargs):
            super().__init__(*args, phase_name=phase_name, tag_phases=tag_phases,
                             phase_folders=phase_folders,
                             optimizer_class=optimizer_class, **kwargs)
            self.optimizer = grid_search.GridSearch(phase_name=phase_name,
                                                    phase_tag=self.phase_tag,
                                                    phase_folders=phase_folders,
                                                    number_of_steps=number_of_steps,
                                                    optimizer_class=optimizer_class,
                                                    model_mapper=self.variable,
                                                    parallel=parallel)

        def run_analysis(self, analysis):
            return self.optimizer.fit(analysis, self.grid_priors)

        # noinspection PyMethodMayBeStatic,PyUnusedLocal
        def make_result(self, result, analysis):
            return result

        @property
        def grid_priors(self):
            raise NotImplementedError(
                "The grid priors property must be implemented to provide a list of "
                "priors to be grid searched")

    return GridSearchExtension



import os


def make_and_return_path_from_path_and_folder_names(path, folder_names):
    """ For a given path, create a directory structure composed of a set of folders and return the path to the \
    inner-most folder.

    For example, if path='/path/to/folders', and folder_names=['folder1', 'folder2'], the directory created will be
    '/path/to/folders/folder1/folder2/' and the returned path will be '/path/to/folders/folder1/folder2/'.

    If the folders already exist, routine continues as normal.

    Parameters
    ----------
    path : str
        The path where the directories are created.
    folder_names : [str]
        The names of the folders which are created in the path directory.

    Returns
    -------
    path
        A string specifying the path to the inner-most folder created.

    Examples
    --------
    path = '/path/to/folders'
    path = make_and_return_path(path=path, folder_names=['folder1', 'folder2'].
    """
    for folder_name in folder_names:

        path += folder_name + '/'

        try:
            os.makedirs(path)
        except FileExistsError:
            pass

    return path



import re

from autofit.mapper.prior import AttributeNameValue


def tuple_name(attribute_name):
    """
    Extract the name of a tuple attribute from the name of one of its components,
    e.g. centre_0 -> origin

    Parameters
    ----------
    attribute_name: str
        The name of an attribute which is a component of a tuple

    Returns
    -------
    tuple_name: str
        The name of the tuple of which the attribute is a member
    """
    return "_".join(attribute_name.split("_")[:-1])


def is_tuple_like_attribute_name(attribute_name):
    """
    Determine if a string matches the pattern "{attribute_name}_#", that is if it
    seems to be a tuple.

    Parameters
    ----------
    attribute_name: str
        The name of some attribute that may refer to a tuple.

    Returns
    -------
    is_tuple_like: bool
        True iff the attribute name looks like that which refers to a tuple.
    """
    pattern = re.compile("^[a-zA-Z_0-9]*_[0-9]$")
    return pattern.match(attribute_name)


class PriorModelNameValue(AttributeNameValue):
    @property
    def prior_model(self):
        return self.value



import numpy as np


def residual_map_from_data_mask_and_model_data(data, mask, model_data):
    """Compute the residual map between a masked observed data and model data, where:

    Residuals = (Data - Model_Data).

    Parameters
    -----------
    data : np.ndarray
        The observed data that is fitted.
    mask : np.ndarray
        The mask applied to the data, where *False* entries are included in the calculation.
    model_data : np.ndarray
        The model data used to fit the observed data.
    """
    return np.subtract(data, model_data, out=np.zeros_like(data), where=np.asarray(mask) == 0)


def normalized_residual_map_from_residual_map_noise_map_and_mask(residual_map, noise_map, mask):
    """Compute the normalized residual map between a masked observed data and model data, where:

    Normalized_Residual = (Data - Model_Data) / Noise

    Parameters
    -----------
    residual_map : np.ndarray
        The residual-map of the model-data fit to the observed data.
    noise_map : np.ndarray
        The noise-map of the observed data.
    mask : np.ndarray
        The mask applied to the residual-map, where *False* entries are included in the calculation.
    """
    return np.divide(residual_map, noise_map, out=np.zeros_like(residual_map), where=np.asarray(mask) == 0)


def chi_squared_map_from_residual_map_noise_map_and_mask(residual_map, noise_map, mask):
    """Computes the chi-squared map between a masked residual-map and noise-map, where:

    Chi_Squared = ((Residuals) / (Noise)) ** 2.0 = ((Data - Model)**2.0)/(Variances)

    Although noise-maps should not contain zero values, it is possible that masking leads to zeros which when \
    divided by create NaNs. Thus, nan_to_num is used to replace these entries with zeros.

    Parameters
    -----------
    residual_map : np.ndarray
        The residual-map of the model-data fit to the observed data.
    noise_map : np.ndarray
        The noise-map of the observed data.
    mask : np.ndarray
        The mask applied to the residual-map, where *False* entries are included in the calculation.
    """
    return np.square(np.divide(residual_map, noise_map, out=np.zeros_like(residual_map),
                               where=np.asarray(mask) == 0))


def chi_squared_from_chi_squared_map_and_mask(chi_squared_map, mask):
    """Compute the chi-squared terms of each model's data-set's fit to an observed data-set, by summing the masked
    chi-squared map of the fit.

    Parameters
    ----------
    chi_squared_map : np.ndarray
        The chi-squared map of values of the model-data fit to the observed data.
    mask : np.ndarray
        The mask applied to the chi-squared map, where *False* entries are included in the calculation.
    """
    return np.sum(chi_squared_map[np.asarray(mask) == 0])


def noise_normalization_from_noise_map_and_mask(noise_map, mask):
    """Compute the noise-map normalization terms of a list of masked 1D noise-maps, summing the noise_map vale in every
    pixel as:

    [Noise_Term] = sum(log(2*pi*[Noise]**2.0))

    Parameters
    ----------
    noise_map : np.ndarray
        The masked noise-map of the observed data.
    mask : np.ndarray
        The mask applied to the noise-map, where *False* entries are included in the calculation.
    """
    return np.sum(np.log(2 * np.pi * noise_map[np.asarray(mask) == 0] ** 2.0))


def likelihood_from_chi_squared_and_noise_normalization(chi_squared, noise_normalization):
    """Compute the likelihood of each masked 1D model-data fit to the data, where:

    Likelihood = -0.5*[Chi_Squared_Term + Noise_Term] (see functions above for these definitions)

    Parameters
    ----------
    chi_squared : float
        The chi-squared term for the model-data fit to the observed data.
    noise_normalization : float
        The normalization noise_map-term for the observed data's noise-map.
    """
    return -0.5 * (chi_squared + noise_normalization)



from . import link
from .model import AbstractModel
from .model import ModelInstance
from .model_mapper import ModelMapper
from .model_mapper import add_to_info_dict
from .model_object import ModelObject
from .prior import ConstantNameValue
from .prior import Prior, UniformPrior, GaussianPrior, LogUniformPrior, \
    AttributeNameValue
from .prior import PriorNameValue
from .prior import cast_collection
from .prior_model import *



import itertools
import os
import shutil

import pytest



import test.mock
import autofit as af




import autofit as af
import autofit as af
import autofit as af
from test.mock import MockClassNLOx4, MockClassNLOx5, MockClassNLOx6, \
    MockNonLinearOptimizer

pytestmark = pytest.mark.filterwarnings('ignore::FutureWarning')

@pytest.fixture(scope="session", autouse=True)
def do_something():
    af.conf.instance = af.conf.Config(
        "{}/../test_files/configs/non_linear".format(os.path.dirname(os.path.realpath(__file__))))


@pytest.fixture(name='mapper')
def make_mapper():
    return af.ModelMapper()


@pytest.fixture(name="mock_list")
def make_mock_list():
    return [af.PriorModel(MockClassNLOx4),
            af.PriorModel(MockClassNLOx4)]


@pytest.fixture(name="result")
def make_result():
    mapper = af.ModelMapper()
    mapper.profile = test.mock.GeometryProfile
    # noinspection PyTypeChecker
    return af.Result(None, None, mapper, [(0, 0), (1, 0)])


class TestResult(object):

    def test_variable(self, result):
        profile = result.variable.profile
        assert profile.centre_0.mean == 0
        assert profile.centre_1.mean == 1
        assert profile.centre_0.sigma == 0.05
        assert profile.centre_1.sigma == 0.05

    def test_variable_absolute(self, result):
        profile = result.variable_absolute(a=2.0).profile
        assert profile.centre_0.mean == 0
        assert profile.centre_1.mean == 1
        assert profile.centre_0.sigma == 2.0
        assert profile.centre_1.sigma == 2.0

    def test_variable_relative(self, result):
        profile = result.variable_relative(r=1.0).profile
        assert profile.centre_0.mean == 0
        assert profile.centre_1.mean == 1
        assert profile.centre_0.sigma == 0.0
        assert profile.centre_1.sigma == 1.0

    def test_raises(self, result):
        with pytest.raises(af.exc.PriorException):
            result.variable.mapper_from_gaussian_tuples(result.gaussian_tuples, a=2.0, r=1.0)


class TestCopyWithNameExtension(object):

    @staticmethod
    def assert_non_linear_attributes_equal(copy, optimizer):
        assert copy.phase_name == "phase_name/one"
        assert copy.variable == optimizer.variable

    def test_copy_with_name_extension(self):
        optimizer = af.NonLinearOptimizer("phase_name")
        copy = optimizer.copy_with_name_extension("one")

        self.assert_non_linear_attributes_equal(copy, optimizer)

    def test_grid_search(self):
        optimizer = af.GridSearch("phase_name", step_size=17, grid=lambda x: x)

        copy = optimizer.copy_with_name_extension("one")
        self.assert_non_linear_attributes_equal(copy, optimizer)
        assert isinstance(copy, af.GridSearch)
        assert copy.step_size is optimizer.step_size
        assert copy.grid is optimizer.grid


# noinspection PyUnresolvedReferences
class TestParamNames(object):

    def test_label_prior_model_tuples(self, mapper, mock_list):
        mapper.mock_list = mock_list

        assert [tup.name for tup in mapper.mock_list.label_prior_model_tuples] == ['0', '1']

    def test_label_prior_model_tuples_with_mapping_name(self, mapper):
        one = af.PriorModel(MockClassNLOx4)
        two = af.PriorModel(MockClassNLOx4)

        one.mapping_name = "one"
        two.mapping_name = "two"

        mapper.mock_list = [one, two]

        assert [tup.name for tup in mapper.mock_list.label_prior_model_tuples] == ['one', 'two']

    def test_prior_prior_model_name_dict(self, mapper, mock_list):
        mapper.mock_list = mock_list
        prior_prior_model_name_dict = mapper.prior_prior_model_name_dict

        assert len({value for key, value in prior_prior_model_name_dict.items()}) == 2


@pytest.fixture(name='nlo_setup_path')
def test_nlo_setup():
    nlo_setup_path = "{}/../test_files/non_linear/nlo/setup/".format(os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(nlo_setup_path):
        shutil.rmtree(nlo_setup_path)

    os.mkdir(nlo_setup_path)

    return nlo_setup_path


@pytest.fixture(name='nlo_model_info_path')
def test_nlo_model_info():
    nlo_model_info_path = "{}/../test_files/non_linear/nlo/model_info/".format(
        os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(nlo_model_info_path):
        shutil.rmtree(nlo_model_info_path)

    return nlo_model_info_path


@pytest.fixture(name='nlo_wrong_info_path')
def test_nlo_wrong_info():
    nlo_wrong_info_path = "{}/../test_files/non_linear/nlo/wrong_info/".format(
        os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(nlo_wrong_info_path):
        shutil.rmtree(nlo_wrong_info_path)

    os.mkdir(nlo_wrong_info_path)

    return nlo_wrong_info_path


class TestDirectorySetup:

    def test__1_class__correct_directory(self, nlo_setup_path):

        af.conf.instance.output_path = nlo_setup_path + '1_class'
        mapper = af.ModelMapper(mock_class=MockClassNLOx4)
        af.NonLinearOptimizer(phase_name='', model_mapper=mapper)

        assert os.path.exists(nlo_setup_path + '1_class')


class TestTotalParameters:

    def test__1_class__four_parameters(self, nlo_setup_path):

        af.conf.instance.output_path = nlo_setup_path + '1_class'
        mapper = af.ModelMapper(mock_class=MockClassNLOx4)
        nlo = af.NonLinearOptimizer(phase_name='', model_mapper=mapper)

        assert nlo.variable.prior_count == 4

    def test__2_classes__six_parameters(self, nlo_setup_path):

        af.conf.instance.output_path = nlo_setup_path + '2_classes'
        mapper = af.ModelMapper(class_1=MockClassNLOx4, class_2=MockClassNLOx6)
        nlo = af.NonLinearOptimizer(phase_name='', model_mapper=mapper)

        assert nlo.variable.prior_count == 10


class TestMostProbableAndLikely(object):

    def test__most_probable_parameters_and_instance__2_classes_6_params(self):

        mapper = af.ModelMapper(mock_class_1=MockClassNLOx4,
                                          mock_class_2=MockClassNLOx6)
        nlo = MockNonLinearOptimizer(phase_name='', model_mapper=mapper,
                                     most_probable=[1.0, 2.0, 3.0, 4.0, -5.0, -6.0, -7.0, -8.0, 9.0, 10.0])

        most_probable = nlo.most_probable_model_instance

        assert most_probable.mock_class_1.one == 1.0
        assert most_probable.mock_class_1.two == 2.0
        assert most_probable.mock_class_1.three == 3.0
        assert most_probable.mock_class_1.four == 4.0

        assert most_probable.mock_class_2.one == (-5.0, -6.0)
        assert most_probable.mock_class_2.two == (-7.0, -8.0)
        assert most_probable.mock_class_2.three == 9.0
        assert most_probable.mock_class_2.four == 10.0

    def test__most_probable__setup_model_instance__1_class_5_params_but_1_is_constant(self):

        mapper = af.ModelMapper(mock_class=MockClassNLOx5)
        mapper.mock_class.five = 10.0

        nlo = MockNonLinearOptimizer(phase_name='', model_mapper=mapper, most_probable=[1.0, -2.0, 3.0, 4.0, 10.0])

        most_probable = nlo.most_probable_model_instance

        assert most_probable.mock_class.one == 1.0
        assert most_probable.mock_class.two == -2.0
        assert most_probable.mock_class.three == 3.0
        assert most_probable.mock_class.four == 4.0
        assert most_probable.mock_class.five == 10.0

    def test__most_likely_parameters_and_instance__2_classes_6_params(self):

        mapper = af.ModelMapper(mock_class_1=MockClassNLOx4,
                                          mock_class_2=MockClassNLOx6)
        nlo = MockNonLinearOptimizer(phase_name='', model_mapper=mapper,
                                     most_likely=[21.0, 22.0, 23.0, 24.0, 25.0, -26.0, -27.0, 28.0, 29.0, 30.0])

        most_likely = nlo.most_likely_model_instance

        assert most_likely.mock_class_1.one == 21.0
        assert most_likely.mock_class_1.two == 22.0
        assert most_likely.mock_class_1.three == 23.0
        assert most_likely.mock_class_1.four == 24.0

        assert most_likely.mock_class_2.one == (25.0, -26.0)
        assert most_likely.mock_class_2.two == (-27.0, 28.0)
        assert most_likely.mock_class_2.three == 29.0
        assert most_likely.mock_class_2.four == 30.0

    def test__most_likely__setup_model_instance__1_class_5_params_but_1_is_constant(self):

        mapper = af.ModelMapper(mock_class=MockClassNLOx5)
        mapper.mock_class.five = 10.0
        nlo = MockNonLinearOptimizer(phase_name='', model_mapper=mapper, most_likely=[9.0, -10.0, -11.0, 12.0, 10.0])

        most_likely = nlo.most_likely_model_instance

        assert most_likely.mock_class.one == 9.0
        assert most_likely.mock_class.two == -10.0
        assert most_likely.mock_class.three == -11.0
        assert most_likely.mock_class.four == 12.0
        assert most_likely.mock_class.five == 10.0


class TestGaussianPriors(object):

    def test__1_class__gaussian_priors_at_3_sigma_confidence(self):

        mapper = af.ModelMapper(mock_class=MockClassNLOx4)
        nlo = MockNonLinearOptimizer(phase_name='', model_mapper=mapper, most_probable=[1.0, 2.0, 3.0, 4.1],
                                     model_lower_params=[0.88, 1.88, 2.88, 3.88],
                                     model_upper_params=[1.12, 2.12, 3.12, 4.12])

        gaussian_priors = nlo.gaussian_priors_at_sigma_limit(sigma_limit=3.0)

        assert gaussian_priors[0][0] == 1.0
        assert gaussian_priors[1][0] == 2.0
        assert gaussian_priors[2][0] == 3.0
        assert gaussian_priors[3][0] == 4.1

        assert gaussian_priors[0][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[1][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[2][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[3][1] == pytest.approx(0.22, 1e-2)


class TestOffsetFromInput:

    def test__input_model_offset_from_most_probable__parameters_and_instance__1_class_4_params(self):

        mapper = af.ModelMapper(mock_class=MockClassNLOx4)
        nlo = MockNonLinearOptimizer(phase_name='', model_mapper=mapper,
                                     most_probable=[1.0, -2.0, 3.0, 4.0])


        offset_values = nlo.offset_values_from_input_model_parameters(input_model_parameters=[1.0, 1.0, 2.0, 3.0])

        assert offset_values == [0.0, -3.0, 1.0, 1.0]

        mapper = af.ModelMapper(mock_class_1=MockClassNLOx4,
                                          mock_class_2=MockClassNLOx6)
        nlo = MockNonLinearOptimizer(phase_name='', model_mapper=mapper,
                                     most_probable=[1.0, 2.0, 3.0, 4.0, -5.0, -6.0, -7.0, -8.0, 9.0, 10.0])

        offset_values = nlo.offset_values_from_input_model_parameters(
            input_model_parameters=[1.0, 1.0, 2.0, 3.0, 10.0, 10.0, 10.0, 10.0, 10.0, 20.0])

        assert offset_values == [0.0, 1.0, 1.0, 1.0, -15.0, -16.0, -17.0, -18.0, -1.0, -10.0]


@pytest.fixture(name='optimizer')
def make_optimizer():
    return af.NonLinearOptimizer(phase_name='', )


class TestLabels(object):

    def test_param_names(self, optimizer):
        optimizer.variable.prior_model = MockClassNLOx4
        assert ['prior_model_one', 'prior_model_two', 'prior_model_three',
                'prior_model_four'] == optimizer.variable.param_names

    def test_properties(self, optimizer):
        optimizer.variable.prior_model = MockClassNLOx4

        assert len(optimizer.param_labels) == 4
        assert len(optimizer.variable.param_names) == 4

    def test_label_config(self):
        assert af.conf.instance.label.label("one") == "x4p0"
        assert af.conf.instance.label.label("two") == "x4p1"
        assert af.conf.instance.label.label("three") == "x4p2"
        assert af.conf.instance.label.label("four") == "x4p3"

    def test_labels(self, optimizer):
        af.AbstractPriorModel._ids = itertools.count()
        optimizer.variable.prior_model = MockClassNLOx4

        assert optimizer.param_labels == [r'x4p0_{\mathrm{a2}}', r'x4p1_{\mathrm{a2}}',
                                          r'x4p2_{\mathrm{a2}}', r'x4p3_{\mathrm{a2}}']


# class TestLatex(object):
#
#    def test__results_at_sigma_limit(self):
#
#        mapper = af.ModelMapper(mock_class=MockClassNLOx4)
#        nlo = MockNonLinearOptimizer(phase_name='', model_mapper=mapper, most_probable=[1.0, 2.0, 3.0, 4.0],
#                                     model_lower_params=[0.5, 1.5, 2.5, 3.5],
#                                     model_upper_params=[1.5, 2.5, 3.5, 4.5])
#
#        latex = nlo.latex_results_at_sigma_limit(sigma_limit=3.0)
#
#        i = 0
#        assert latex[i] == r'x4p0_{\mathrm{a2}} = 1.00^{+1.50}_{-0.50} & ' ; i+=1
#        assert latex[i] == r'x4p1_{\mathrm{a2}} = 2.00^{+2.50}_{-1.50} & ' ; i+=1
#        assert latex[i] == r'x4p2_{\mathrm{a2}} = 3.00^{+3.50}_{-2.50} & ' ; i+=1
#        assert latex[i] == r'x4p3_{\mathrm{a2}} = 4.00^{+4.50}_{-3.50} & ' ; i+=1